
/*============================================================================

(c) Don Cross <dcross@intersrv.com>
Modified by Vadim Engelson, <vadim.engelson@mathcore.com>
============================================================================*/

#include "Global.h"
// a package name where this code is integrated to

#define TRUE  true
#define FALSE false

#define BITS_PER_WORD   (sizeof(unsigned) * 8)


int IsPowerOfTwo ( unsigned x )
{
    if ( x < 2 )
        return FALSE;

    if ( x & (x-1) )        // Thanks to 'byang' for this cute trick!
        return FALSE;

    return TRUE;
}


unsigned NumberOfBitsNeeded ( unsigned PowerOfTwo )
{
    unsigned i;

    if ( PowerOfTwo < 2 )
    {

    }

    for ( i=0; ; i++ )
    {
        if ( PowerOfTwo & (1 << i) )
            return i;
    }
}



unsigned ReverseBits ( unsigned index, unsigned NumBits )
{
    unsigned i, rev;

    for ( i=rev=0; i < NumBits; i++ )
    {
        rev = (rev << 1) | (index & 1);
        index >>= 1;
    }

    return rev;
}


double Index_to_frequency ( unsigned NumSamples, unsigned Index )
{
    if ( Index >= NumSamples )
        return 0.0;
    else if ( Index <= NumSamples/2 )
        return (double)Index / (double)NumSamples;

    return -(double)(NumSamples-Index) / (double)NumSamples;
}






#define CHECKPOINTER(p) 



int fft_double (
                unsigned  NumSamples,
                int       InverseTransform,
                double   *RealIn,
                double   *ImagIn,
                double   *RealOut,
                double   *ImagOut )
{
    unsigned NumBits;    /* Number of bits needed to store indices */
    unsigned i, j, k, n;
    unsigned BlockSize, BlockEnd;

    double angle_numerator = 2.0 * 3.14159265358979323846;
    double tr, ti;     /* temp real, temp imaginary */

    if ( !IsPowerOfTwo(NumSamples) )
    {
        return 1;
    }

    if ( InverseTransform )
        angle_numerator = -angle_numerator;

    CHECKPOINTER ( RealIn );
    CHECKPOINTER ( RealOut );
    CHECKPOINTER ( ImagOut );

    NumBits = NumberOfBitsNeeded ( NumSamples );

    /*
    **   Do simultaneous data copy and bit-reversal ordering into outputs...
    */

    for ( i=0; i < NumSamples; i++ )
    {
        j = ReverseBits ( i, NumBits );
        RealOut[j] = RealIn[i];
        ImagOut[j] = (ImagIn == NULL) ? 0.0 : ImagIn[i];
    }

    /*
    **   Do the FFT itself...
    */

    BlockEnd = 1;
    for ( BlockSize = 2; BlockSize <= NumSamples; BlockSize <<= 1 )
    {
        double delta_angle = angle_numerator / (double)BlockSize;
        double sm2 = sin ( -2 * delta_angle );
        double sm1 = sin ( -delta_angle );
        double cm2 = cos ( -2 * delta_angle );
        double cm1 = cos ( -delta_angle );
        double w = 2 * cm1;
        double ar[3], ai[3];
        /*double temp;*/

        for ( i=0; i < NumSamples; i += BlockSize )
        {
            ar[2] = cm2;
            ar[1] = cm1;

            ai[2] = sm2;
            ai[1] = sm1;

            for ( j=i, n=0; n < BlockEnd; j++, n++ )
            {
                ar[0] = w*ar[1] - ar[2];
                ar[2] = ar[1];
                ar[1] = ar[0];

                ai[0] = w*ai[1] - ai[2];
                ai[2] = ai[1];
                ai[1] = ai[0];

                k = j + BlockEnd;
                tr = ar[0]*RealOut[k] - ai[0]*ImagOut[k];
                ti = ar[0]*ImagOut[k] + ai[0]*RealOut[k];

                RealOut[k] = RealOut[j] - tr;
                ImagOut[k] = ImagOut[j] - ti;

                RealOut[j] += tr;
                ImagOut[j] += ti;
            }
        }

        BlockEnd = BlockSize;
    }

    /*
    **   Need to normalize if inverse transform...
    */

    if ( InverseTransform )
    {
        double denom = (double)NumSamples;

        for ( i=0; i < NumSamples; i++ )
        {
            RealOut[i] /= denom;
            ImagOut[i] /= denom;
        }
    }
    return 0;
}

lm_complexN FRadix(
                   /* Input  */ const lm_complexN  & in,
                   /* Input  */ const int  & inv,
                   /* Output  */ int  & mcO1)
{


    unsigned  NumSamples=in.dimension(1);
    double   *RealIn=new double[NumSamples];
    double   *ImagIn=new double[NumSamples];
    double   *RealOut=new double[NumSamples];
    double   *ImagOut=new double[NumSamples];
    lm_complexN ret=in;

    int i;
    for (i=0;i<NumSamples;i++) {
        RealIn[i]=in(i+1).re;
        ImagIn[i]=in(i+1).im;
    };

    int res=fft_double (
        NumSamples,
        inv,
        RealIn,
        ImagIn,
        RealOut,
        ImagOut );


    if (res==0){

        for (i=0;i<NumSamples;i++) {
            ret(i+1).re=RealOut[i];
            ret(i+1).im=ImagOut[i];
        };

        if (inv) {
            ret=ret*sqrt((double)NumSamples);
            /* otherwise sqrt() is ambiguous */
        } 		
        else 
        {
            ret=ret/sqrt((double)NumSamples);
            /* otherwise sqrt() is ambiguous */
        }

    }
    else
    {
        ret=lm_complex(-1,-1);
    }
    delete [] RealIn;
    delete [] ImagIn;
    delete [] RealOut;
    delete [] ImagOut;

    mcO1=res;
    return ret;

}




/*--- end of file fourierd.c ---*/
