(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[     97818,       2768]
NotebookOptionsPosition[     55611,       1795]
NotebookOutlinePosition[     92869,       2609]
CellTagsIndexPosition[     92826,       2606]
WindowFrame->Normal
ContainsDynamic->False*)

(* Beginning of Notebook Content *)
Notebook[{
Cell["Code Generation for a Symbolic Matrix", "Title"],

Cell["\<\
Copyright \[Copyright] MathCore AB 1999-2011\
\>", "Copyright",
 CellChangeTimes->{{3.5075244762416086`*^9, 3.507524478710327*^9}}],

Cell[TextData[{
 "In this notebook we give an example of how ",
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 " can be used to generate C++ code for efficient computation of a large \
matrix with symbolic entries. The matrix in this case is a Hessian of a given \
function which is an important entity to compute efficiently in for example \
optimization algorithms."
}], "Text"],

Cell[CellGroupData[{

Cell["A Small Exampe", "Section"],

Cell[CellGroupData[{

Cell["An Expression", "Subsection"],

Cell["\<\
We will use the following expression for generating a large matrix with \
symbolic entries.\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"expr", "=", 
   FractionBox[
    RowBox[{"(", 
     RowBox[{
      RowBox[{"Sin", "[", 
       RowBox[{"y", " ", "z"}], "]"}], "+", 
      RowBox[{"ArcTan", "[", "x", "]"}]}], ")"}], 
    SqrtBox[
     RowBox[{"1", "+", 
      RowBox[{".5", 
       RowBox[{"\[ExponentialE]", "^", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{"x", "^", "2"}]}], "-", 
          RowBox[{"y", "^", "2"}]}], ")"}]}]}]}]]]}], ";"}]], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["A Function for Computing the Gradient", "Subsection"],

Cell[TextData[{
 StyleBox["Grad",
  FontWeight->"Bold"],
 " computes the gradient of a function w.r.t. a list of variables "
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Off", "[", 
   RowBox[{"General", "::", "spell1"}], "]"}], ";", " ", 
  RowBox[{"(*", 
   RowBox[{"To", " ", "avoid", " ", "warnings"}], "*)"}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Grad", ";", 
   RowBox[{"On", "[", 
    RowBox[{"General", "::", "spell1"}], "]"}], ";"}], 
  RowBox[{"(*", 
   RowBox[{"To", " ", "avoid", " ", "warnings"}], "*)"}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Grad", "[", 
   RowBox[{"fcn_", ",", "vars_"}], "]"}], ":=", 
  RowBox[{"Map", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      SubscriptBox["\[PartialD]", "#"], "fcn"}], "&"}], ",", "vars"}], 
   "]"}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Grad", "[", 
  RowBox[{"expr", ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "y", ",", "z"}], "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    FractionBox["1", 
     RowBox[{
      SqrtBox[
       RowBox[{"1", "+", 
        RowBox[{"0.5`", " ", 
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox["x", "2"]}], "-", 
           SuperscriptBox["y", "2"]}]]}]}]], " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", 
        SuperscriptBox["x", "2"]}], ")"}]}]], "+", 
    FractionBox[
     RowBox[{"0.5`", " ", 
      SuperscriptBox["\[ExponentialE]", 
       RowBox[{
        RowBox[{"-", 
         SuperscriptBox["x", "2"]}], "-", 
        SuperscriptBox["y", "2"]}]], " ", "x", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"ArcTan", "[", "x", "]"}], "+", 
        RowBox[{"Sin", "[", 
         RowBox[{"y", " ", "z"}], "]"}]}], ")"}]}], 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", 
        RowBox[{"0.5`", " ", 
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox["x", "2"]}], "-", 
           SuperscriptBox["y", "2"]}]]}]}], ")"}], 
      RowBox[{"3", "/", "2"}]]]}], ",", 
   RowBox[{
    FractionBox[
     RowBox[{"z", " ", 
      RowBox[{"Cos", "[", 
       RowBox[{"y", " ", "z"}], "]"}]}], 
     SqrtBox[
      RowBox[{"1", "+", 
       RowBox[{"0.5`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]]}]}]]], "+", 
    FractionBox[
     RowBox[{"0.5`", " ", 
      SuperscriptBox["\[ExponentialE]", 
       RowBox[{
        RowBox[{"-", 
         SuperscriptBox["x", "2"]}], "-", 
        SuperscriptBox["y", "2"]}]], " ", "y", " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"ArcTan", "[", "x", "]"}], "+", 
        RowBox[{"Sin", "[", 
         RowBox[{"y", " ", "z"}], "]"}]}], ")"}]}], 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"1", "+", 
        RowBox[{"0.5`", " ", 
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{
           RowBox[{"-", 
            SuperscriptBox["x", "2"]}], "-", 
           SuperscriptBox["y", "2"]}]]}]}], ")"}], 
      RowBox[{"3", "/", "2"}]]]}], ",", 
   RowBox[{"0.`", "\[VeryThinSpace]", "+", 
    FractionBox[
     RowBox[{"y", " ", 
      RowBox[{"Cos", "[", 
       RowBox[{"y", " ", "z"}], "]"}]}], 
     SqrtBox[
      RowBox[{"1", "+", 
       RowBox[{"0.5`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]]}]}]]]}]}], "}"}]], "Output",
 CellChangeTimes->{3.5003728438551774`*^9, 3.5003729380096197`*^9, 
  3.500372973508938*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["A Function for Computing the Hessian", "Subsection"],

Cell[TextData[{
 StyleBox["Hessian",
  FontWeight->"Bold"],
 " computes the Hessian matrix of a function w.r.t. a list of variables. \
HessianMC should be used because Hessian[] already exists in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ". "
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"HessianMC", "[", 
   RowBox[{"fcn_", ",", "vars_"}], "]"}], ":=", 
  RowBox[{"Map", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Grad", "[", 
      RowBox[{"#", ",", "vars"}], "]"}], "&"}], ",", 
    RowBox[{"Grad", "[", 
     RowBox[{"fcn", ",", "vars"}], "]"}]}], "]"}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"HessianMC", "[", 
  RowBox[{"expr", ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "y", ",", "z"}], "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       FractionBox[
        RowBox[{"2", " ", "x"}], 
        RowBox[{
         SqrtBox[
          RowBox[{"1", "+", 
           RowBox[{"0.5`", " ", 
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{
              RowBox[{"-", 
               SuperscriptBox["x", "2"]}], "-", 
              SuperscriptBox["y", "2"]}]]}]}]], " ", 
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{"1", "+", 
            SuperscriptBox["x", "2"]}], ")"}], "2"]}]]}], "+", 
      FractionBox[
       RowBox[{"1.`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]], " ", "x"}], 
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", 
           RowBox[{"0.5`", " ", 
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{
              RowBox[{"-", 
               SuperscriptBox["x", "2"]}], "-", 
              SuperscriptBox["y", "2"]}]]}]}], ")"}], 
         RowBox[{"3", "/", "2"}]], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", 
          SuperscriptBox["x", "2"]}], ")"}]}]], "+", 
      FractionBox[
       RowBox[{"0.5`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"ArcTan", "[", "x", "]"}], "+", 
          RowBox[{"Sin", "[", 
           RowBox[{"y", " ", "z"}], "]"}]}], ")"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"0.5`", " ", 
           SuperscriptBox["\[ExponentialE]", 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["x", "2"]}], "-", 
             SuperscriptBox["y", "2"]}]]}]}], ")"}], 
        RowBox[{"3", "/", "2"}]]], "+", 
      FractionBox[
       RowBox[{"0.75`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", 
           SuperscriptBox["x", "2"]}], "-", 
          RowBox[{"2", " ", 
           SuperscriptBox["y", "2"]}]}]], " ", 
        SuperscriptBox["x", "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"ArcTan", "[", "x", "]"}], "+", 
          RowBox[{"Sin", "[", 
           RowBox[{"y", " ", "z"}], "]"}]}], ")"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"0.5`", " ", 
           SuperscriptBox["\[ExponentialE]", 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["x", "2"]}], "-", 
             SuperscriptBox["y", "2"]}]]}]}], ")"}], 
        RowBox[{"5", "/", "2"}]]], "-", 
      FractionBox[
       RowBox[{"1.`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]], " ", 
        SuperscriptBox["x", "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"ArcTan", "[", "x", "]"}], "+", 
          RowBox[{"Sin", "[", 
           RowBox[{"y", " ", "z"}], "]"}]}], ")"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"0.5`", " ", 
           SuperscriptBox["\[ExponentialE]", 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["x", "2"]}], "-", 
             SuperscriptBox["y", "2"]}]]}]}], ")"}], 
        RowBox[{"3", "/", "2"}]]]}], ",", 
     RowBox[{
      FractionBox[
       RowBox[{"0.5`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]], " ", "y"}], 
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", 
           RowBox[{"0.5`", " ", 
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{
              RowBox[{"-", 
               SuperscriptBox["x", "2"]}], "-", 
              SuperscriptBox["y", "2"]}]]}]}], ")"}], 
         RowBox[{"3", "/", "2"}]], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", 
          SuperscriptBox["x", "2"]}], ")"}]}]], "+", 
      FractionBox[
       RowBox[{"0.5`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]], " ", "x", " ", "z", " ", 
        RowBox[{"Cos", "[", 
         RowBox[{"y", " ", "z"}], "]"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"0.5`", " ", 
           SuperscriptBox["\[ExponentialE]", 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["x", "2"]}], "-", 
             SuperscriptBox["y", "2"]}]]}]}], ")"}], 
        RowBox[{"3", "/", "2"}]]], "+", 
      FractionBox[
       RowBox[{"0.75`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", 
           SuperscriptBox["x", "2"]}], "-", 
          RowBox[{"2", " ", 
           SuperscriptBox["y", "2"]}]}]], " ", "x", " ", "y", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"ArcTan", "[", "x", "]"}], "+", 
          RowBox[{"Sin", "[", 
           RowBox[{"y", " ", "z"}], "]"}]}], ")"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"0.5`", " ", 
           SuperscriptBox["\[ExponentialE]", 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["x", "2"]}], "-", 
             SuperscriptBox["y", "2"]}]]}]}], ")"}], 
        RowBox[{"5", "/", "2"}]]], "-", 
      FractionBox[
       RowBox[{"1.`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]], " ", "x", " ", "y", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"ArcTan", "[", "x", "]"}], "+", 
          RowBox[{"Sin", "[", 
           RowBox[{"y", " ", "z"}], "]"}]}], ")"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"0.5`", " ", 
           SuperscriptBox["\[ExponentialE]", 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["x", "2"]}], "-", 
             SuperscriptBox["y", "2"]}]]}]}], ")"}], 
        RowBox[{"3", "/", "2"}]]]}], ",", 
     RowBox[{"0.`", "\[VeryThinSpace]", "+", 
      FractionBox[
       RowBox[{"0.5`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]], " ", "x", " ", "y", " ", 
        RowBox[{"Cos", "[", 
         RowBox[{"y", " ", "z"}], "]"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"0.5`", " ", 
           SuperscriptBox["\[ExponentialE]", 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["x", "2"]}], "-", 
             SuperscriptBox["y", "2"]}]]}]}], ")"}], 
        RowBox[{"3", "/", "2"}]]]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      FractionBox[
       RowBox[{"0.5`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]], " ", "y"}], 
       RowBox[{
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"1", "+", 
           RowBox[{"0.5`", " ", 
            SuperscriptBox["\[ExponentialE]", 
             RowBox[{
              RowBox[{"-", 
               SuperscriptBox["x", "2"]}], "-", 
              SuperscriptBox["y", "2"]}]]}]}], ")"}], 
         RowBox[{"3", "/", "2"}]], " ", 
        RowBox[{"(", 
         RowBox[{"1", "+", 
          SuperscriptBox["x", "2"]}], ")"}]}]], "+", 
      FractionBox[
       RowBox[{"0.5`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]], " ", "x", " ", "z", " ", 
        RowBox[{"Cos", "[", 
         RowBox[{"y", " ", "z"}], "]"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"0.5`", " ", 
           SuperscriptBox["\[ExponentialE]", 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["x", "2"]}], "-", 
             SuperscriptBox["y", "2"]}]]}]}], ")"}], 
        RowBox[{"3", "/", "2"}]]], "+", 
      FractionBox[
       RowBox[{"0.75`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", 
           SuperscriptBox["x", "2"]}], "-", 
          RowBox[{"2", " ", 
           SuperscriptBox["y", "2"]}]}]], " ", "x", " ", "y", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"ArcTan", "[", "x", "]"}], "+", 
          RowBox[{"Sin", "[", 
           RowBox[{"y", " ", "z"}], "]"}]}], ")"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"0.5`", " ", 
           SuperscriptBox["\[ExponentialE]", 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["x", "2"]}], "-", 
             SuperscriptBox["y", "2"]}]]}]}], ")"}], 
        RowBox[{"5", "/", "2"}]]], "-", 
      FractionBox[
       RowBox[{"1.`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]], " ", "x", " ", "y", " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"ArcTan", "[", "x", "]"}], "+", 
          RowBox[{"Sin", "[", 
           RowBox[{"y", " ", "z"}], "]"}]}], ")"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"0.5`", " ", 
           SuperscriptBox["\[ExponentialE]", 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["x", "2"]}], "-", 
             SuperscriptBox["y", "2"]}]]}]}], ")"}], 
        RowBox[{"3", "/", "2"}]]]}], ",", 
     RowBox[{
      FractionBox[
       RowBox[{"1.`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]], " ", "y", " ", "z", " ", 
        RowBox[{"Cos", "[", 
         RowBox[{"y", " ", "z"}], "]"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"0.5`", " ", 
           SuperscriptBox["\[ExponentialE]", 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["x", "2"]}], "-", 
             SuperscriptBox["y", "2"]}]]}]}], ")"}], 
        RowBox[{"3", "/", "2"}]]], "-", 
      FractionBox[
       RowBox[{
        SuperscriptBox["z", "2"], " ", 
        RowBox[{"Sin", "[", 
         RowBox[{"y", " ", "z"}], "]"}]}], 
       SqrtBox[
        RowBox[{"1", "+", 
         RowBox[{"0.5`", " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{
            RowBox[{"-", 
             SuperscriptBox["x", "2"]}], "-", 
            SuperscriptBox["y", "2"]}]]}]}]]], "+", 
      FractionBox[
       RowBox[{"0.5`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"ArcTan", "[", "x", "]"}], "+", 
          RowBox[{"Sin", "[", 
           RowBox[{"y", " ", "z"}], "]"}]}], ")"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"0.5`", " ", 
           SuperscriptBox["\[ExponentialE]", 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["x", "2"]}], "-", 
             SuperscriptBox["y", "2"]}]]}]}], ")"}], 
        RowBox[{"3", "/", "2"}]]], "+", 
      FractionBox[
       RowBox[{"0.75`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{
           RowBox[{"-", "2"}], " ", 
           SuperscriptBox["x", "2"]}], "-", 
          RowBox[{"2", " ", 
           SuperscriptBox["y", "2"]}]}]], " ", 
        SuperscriptBox["y", "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"ArcTan", "[", "x", "]"}], "+", 
          RowBox[{"Sin", "[", 
           RowBox[{"y", " ", "z"}], "]"}]}], ")"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"0.5`", " ", 
           SuperscriptBox["\[ExponentialE]", 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["x", "2"]}], "-", 
             SuperscriptBox["y", "2"]}]]}]}], ")"}], 
        RowBox[{"5", "/", "2"}]]], "-", 
      FractionBox[
       RowBox[{"1.`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]], " ", 
        SuperscriptBox["y", "2"], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"ArcTan", "[", "x", "]"}], "+", 
          RowBox[{"Sin", "[", 
           RowBox[{"y", " ", "z"}], "]"}]}], ")"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"0.5`", " ", 
           SuperscriptBox["\[ExponentialE]", 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["x", "2"]}], "-", 
             SuperscriptBox["y", "2"]}]]}]}], ")"}], 
        RowBox[{"3", "/", "2"}]]]}], ",", 
     RowBox[{"0.`", "\[VeryThinSpace]", "+", 
      FractionBox[
       RowBox[{"Cos", "[", 
        RowBox[{"y", " ", "z"}], "]"}], 
       SqrtBox[
        RowBox[{"1", "+", 
         RowBox[{"0.5`", " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{
            RowBox[{"-", 
             SuperscriptBox["x", "2"]}], "-", 
            SuperscriptBox["y", "2"]}]]}]}]]], "+", 
      FractionBox[
       RowBox[{"0.5`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]], " ", 
        SuperscriptBox["y", "2"], " ", 
        RowBox[{"Cos", "[", 
         RowBox[{"y", " ", "z"}], "]"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"0.5`", " ", 
           SuperscriptBox["\[ExponentialE]", 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["x", "2"]}], "-", 
             SuperscriptBox["y", "2"]}]]}]}], ")"}], 
        RowBox[{"3", "/", "2"}]]], "-", 
      FractionBox[
       RowBox[{"y", " ", "z", " ", 
        RowBox[{"Sin", "[", 
         RowBox[{"y", " ", "z"}], "]"}]}], 
       SqrtBox[
        RowBox[{"1", "+", 
         RowBox[{"0.5`", " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{
            RowBox[{"-", 
             SuperscriptBox["x", "2"]}], "-", 
            SuperscriptBox["y", "2"]}]]}]}]]]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     FractionBox[
      RowBox[{"0.5`", " ", 
       SuperscriptBox["\[ExponentialE]", 
        RowBox[{
         RowBox[{"-", 
          SuperscriptBox["x", "2"]}], "-", 
         SuperscriptBox["y", "2"]}]], " ", "x", " ", "y", " ", 
       RowBox[{"Cos", "[", 
        RowBox[{"y", " ", "z"}], "]"}]}], 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{"1", "+", 
         RowBox[{"0.5`", " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{
            RowBox[{"-", 
             SuperscriptBox["x", "2"]}], "-", 
            SuperscriptBox["y", "2"]}]]}]}], ")"}], 
       RowBox[{"3", "/", "2"}]]], ",", 
     RowBox[{
      FractionBox[
       RowBox[{"Cos", "[", 
        RowBox[{"y", " ", "z"}], "]"}], 
       SqrtBox[
        RowBox[{"1", "+", 
         RowBox[{"0.5`", " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{
            RowBox[{"-", 
             SuperscriptBox["x", "2"]}], "-", 
            SuperscriptBox["y", "2"]}]]}]}]]], "+", 
      FractionBox[
       RowBox[{"0.5`", " ", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]], " ", 
        SuperscriptBox["y", "2"], " ", 
        RowBox[{"Cos", "[", 
         RowBox[{"y", " ", "z"}], "]"}]}], 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"1", "+", 
          RowBox[{"0.5`", " ", 
           SuperscriptBox["\[ExponentialE]", 
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["x", "2"]}], "-", 
             SuperscriptBox["y", "2"]}]]}]}], ")"}], 
        RowBox[{"3", "/", "2"}]]], "-", 
      FractionBox[
       RowBox[{"y", " ", "z", " ", 
        RowBox[{"Sin", "[", 
         RowBox[{"y", " ", "z"}], "]"}]}], 
       SqrtBox[
        RowBox[{"1", "+", 
         RowBox[{"0.5`", " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{
            RowBox[{"-", 
             SuperscriptBox["x", "2"]}], "-", 
            SuperscriptBox["y", "2"]}]]}]}]]]}], ",", 
     RowBox[{"0.`", "\[VeryThinSpace]", "-", 
      FractionBox[
       RowBox[{
        SuperscriptBox["y", "2"], " ", 
        RowBox[{"Sin", "[", 
         RowBox[{"y", " ", "z"}], "]"}]}], 
       SqrtBox[
        RowBox[{"1", "+", 
         RowBox[{"0.5`", " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{
            RowBox[{"-", 
             SuperscriptBox["x", "2"]}], "-", 
            SuperscriptBox["y", "2"]}]]}]}]]]}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.500372843902052*^9, 3.5003729380721183`*^9, 
  3.5003729735401874`*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["A Symbolic Matrix", "Subsection"],

Cell[TextData[{
 "We define a function that evaluates the Hessian of a given expression ",
 StyleBox["expr",
  FontWeight->"Bold"]
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"SymbolicMatrix", "[", 
    RowBox[{"x_", ",", "y_", ",", "z_"}], "]"}], "=", 
   RowBox[{"HessianMC", "[", 
    RowBox[{"expr", ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", "y", ",", "z"}], "}"}]}], "]"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.507524464663632*^9, 3.507524465444872*^9}}],

Cell["A numerical evaluation", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SymbolicMatrix", "[", 
   RowBox[{"1.", ",", "2.", ",", "3."}], "]"}], "//", "MatrixForm"}]], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"-", "0.4974868384054093`"}], "0.006257462703369232`", 
      "0.006437019895222956`"},
     {"0.006257462703369232`", "2.537332121026991`", "2.6451070821495954`"},
     {"0.006437019895222956`", "2.6451070821495954`", "1.115784049703575`"}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.500372843980175*^9, 3.500372938150242*^9, 
  3.500372973602686*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Code Generation", "Subsection"],

Cell[TextData[{
 "Load the ",
 StyleBox["MathCode C++",
  FontSlant->"Italic"],
 " application package"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<MathCode`\>\"", "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"MathCode C++ 1.4.4\"\>", "\[InvisibleSpace]", "\<\" for \"\>", 
   "\[InvisibleSpace]", "\<\"mingw32\"\>", 
   "\[InvisibleSpace]", "\<\" loaded from \"\>", 
   "\[InvisibleSpace]", "\<\"C:\\\\b\\\\omath\\\\MathCode\"\>"}],
  SequenceForm[
  "MathCode C++ 1.4.4", " for ", "mingw32", " loaded from ", 
   "C:\\b\\omath\\MathCode"],
  Editable->False]], "Print",
 CellChangeTimes->{3.500372938556484*^9, 3.5003729739933033`*^9}]
}, Open  ]],

Cell["SetDirectory[$MCRoot<>\"/Demos/CompiledHessian\"];", "Input"],

Cell[TextData[{
 "We need to declare types of the argument and return value of the function \
to be compiled. In this case ",
 "3",
 " real parameters and a real 3",
 StyleBox[" \[Cross] ",
  FontWeight->"Bold"],
 "3 matrix."
}], "Text"],

Cell[BoxData[
 RowBox[{"Declare", "[", 
  RowBox[{
   RowBox[{"SymbolicMatrix", "[", 
    RowBox[{
     RowBox[{"Real", " ", "x_"}], ",", 
     RowBox[{"Real", " ", "y_"}], ",", 
     RowBox[{"Real", " ", "z_"}]}], "]"}], "\[Rule]", 
   RowBox[{"Real", "[", 
    RowBox[{"3", ",", "3"}], "]"}]}], "]"}]], "Input"],

Cell[TextData[{
 "We generate C++ code and compile it to an executable file using the ",
 StyleBox["BuildCode",
  FontWeight->"Bold"],
 " command. The option ",
 StyleBox["EvaluateFunctions",
  FontWeight->"Bold"],
 " makes optimizations such as common subexpression eliminations and symbolic \
evaluation of the function ",
 StyleBox["SymbolicMatrix",
  FontWeight->"Bold"],
 " before code generation."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BuildCode", "[", 
  RowBox[{"EvaluateFunctions", "\[Rule]", 
   RowBox[{"{", "SymbolicMatrix", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"Grad", ",", "HessianMC"}], "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`Grad, \
$CellContext`HessianMC}],
  Editable->False]], "Print",
 CellChangeTimes->{3.5003728465582504`*^9, 3.5003729389158525`*^9, 
  3.5003729742745485`*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"Grad", ",", "HessianMC"}], "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`Grad, \
$CellContext`HessianMC}],
  Editable->False]], "Print",
 CellChangeTimes->{3.5003728465582504`*^9, 3.5003729389158525`*^9, 
  3.5003729743057976`*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"Grad", ",", "HessianMC"}], "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`Grad, \
$CellContext`HessianMC}],
  Editable->False]], "Print",
 CellChangeTimes->{3.5003728465582504`*^9, 3.5003729389158525`*^9, 
  3.5003729743214226`*^9}],

Cell[BoxData["\<\"Successful compilation to C++: 1 function(s)\"\>"], "Print",
 CellChangeTimes->{3.5003728465582504`*^9, 3.5003729389158525`*^9, 
  3.5003729765401297`*^9}]
}, Open  ]],

Cell[TextData[{
 "Connect the executable to ",
 StyleBox["Mathematica",
  FontSlant->"Italic"]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"InstallCode", "[", "]"}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Global\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Global", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.500372981165041*^9}]
}, Open  ]],

Cell["Now the external code is used instead", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SymbolicMatrix", "[", 
   RowBox[{"1.", ",", "2.", ",", "3."}], "]"}], "//", "MatrixForm"}]], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"-", "0.4974868384054093`"}], "0.006257462703369234`", 
      "0.006437019895222957`"},
     {"0.006257462703369234`", "2.5373321210269904`", "2.6451070821495954`"},
     {"0.006437019895222957`", "2.6451070821495954`", "1.1157840497035747`"}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.5003728490113287`*^9, 3.500372981211915*^9}]
}, Open  ]],

Cell[TextData[{
 "Remove the link to the external code (",
 StyleBox["Mathematicas",
  FontSlant->"Italic"],
 " own definition of ",
 StyleBox["SymbolicMatrix",
  FontWeight->"Bold"],
 " is used again)."
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"UninstallCode", "[", "]"}], ";"}]], "Input"],

Cell[TextData[{
 "Remove the definition of ",
 StyleBox["SymbolicMatrix",
  FontWeight->"Bold"]
}], "Text"],

Cell[BoxData[
 RowBox[{"Remove", "[", "SymbolicMatrix", "]"}]], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["The Generated Code", "Subsection"],

Cell[TextData[{
 "Here follows a listing of the generated code. Note the extensive use of \
temporary variables. The option ",
 StyleBox["EvaluateFunctions",
  FontWeight->"Bold"],
 " used above in the ",
 StyleBox["BuildCode",
  FontWeight->"Bold"],
 " command decomposes the large symbolic expression in such a way that common \
subexpressions only are evaluated once and then stored in temporary variables \
to be reused in the numeric evaluation of the rhs of the function. This gives \
extremely efficient code!"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FilePrint", "[", "\"\<Global.cc\>\"", "]"}]], "Input"],

Cell["\<\
#include \"Global.h\"

#include \"Global.icc\"

#include <math.h>
void Global_TGlobalInit ()
{
; 
}

doubleNN Global_TSymbolicMatrix ( const double &x, const double &y, const 
    double &z)
{
    double mc_T1;
    double mc_T2;
    double mc_T3;
    double mc_T4;
    double mc_T5;
    double mc_T6;
    double mc_T7;
    double mc_T8;
    double mc_T9;
    double mc_T10;
    double mc_T11;
    double mc_T12;
    double mc_T13;
    double mc_T14;
    double mc_T15;
    double mc_T16;
    double mc_T17;
    double mc_T18;
    double mc_T19;
    double mc_T20;
    double mc_T21;
    double mc_T22;
    double mc_T23;
    double mc_T24;
    doubleN mc_T25(3);
    double mc_T26;
    double mc_T27;
    double mc_T28;
    double mc_T29;
    double mc_T30;
    double mc_T31;
    double mc_T32;
    double mc_T33;
    double mc_T34;
    double mc_T35;
    double mc_T36;
    double mc_T37;
    double mc_T38;
    double mc_T39;
    double mc_T40;
    double mc_T41;
    double mc_T42;
    double mc_T43;
    double mc_T44;
    double mc_T45;
    double mc_T46;
    double mc_T47;
    double mc_T48;
    double mc_T49;
    doubleN mc_T50(3);
    double mc_T51;
    double mc_T52;
    double mc_T53;
    double mc_T54;
    double mc_T55;
    double mc_T56;
    double mc_T57;
    double mc_T58;
    doubleN mc_T59(3);
    doubleNN mc_T60(3, 3);
    mc_T1 = y*z;
    mc_T2 = sin(mc_T1);
    mc_T3 = (y*y);
    mc_T4 =                       -0.5;
    mc_T5 = -mc_T3;
    mc_T6 = (x*x);
    mc_T7 = -mc_T6;
    mc_T8 = mc_T7+mc_T5;
    mc_T9 = pow(2.71828182845904523536, mc_T8);
    mc_T10 = 0.5*mc_T9;
    mc_T11 = 1+mc_T10;
    mc_T12 = pow(mc_T11, mc_T4);
    mc_T13 = mc_T12*mc_T3*mc_T2;
    mc_T14 = -mc_T13;
    mc_T15 = 0.+mc_T14;
    mc_T16 = mc_T12*y*z*mc_T2;
    mc_T17 = -mc_T16;
    mc_T18 = cos(mc_T1);
    mc_T19 =                       -1.5;
    mc_T20 = pow(mc_T11, mc_T19);
    mc_T21 = 0.5*mc_T9*mc_T20*mc_T3*mc_T18;
    mc_T22 = mc_T12*mc_T18;
    mc_T23 = mc_T22+mc_T21+mc_T17;
    mc_T24 = 0.5*mc_T9*mc_T20*x*y*mc_T18;
    mc_T25.SetShape(3 );
    mc_T25(1) = mc_T24;
    mc_T25(2) = mc_T23;
    mc_T25(3) = mc_T15;

    mc_T26 = 0.+mc_T22+mc_T21+mc_T17;
    mc_T27 = atan(x);
    mc_T28 = mc_T27+mc_T2;
    mc_T29 = -1.*mc_T9*mc_T20*mc_T3*mc_T28;
    mc_T30 =                       -2.5;
    mc_T31 = pow(mc_T11, mc_T30);
    mc_T32 = -2*mc_T3;
    mc_T33 = -2*mc_T6;
    mc_T34 = mc_T33+mc_T32;
    mc_T35 = pow(2.71828182845904523536, mc_T34);
    mc_T36 = 0.75*mc_T35*mc_T31*mc_T3*mc_T28;
    mc_T37 = 0.5*mc_T9*mc_T20*mc_T28;
    mc_T38 = (z*z);
    mc_T39 = mc_T12*mc_T38*mc_T2;
    mc_T40 = -mc_T39;
    mc_T41 = 1.*mc_T9*mc_T20*y*z*mc_T18;
    mc_T42 = mc_T41+mc_T40+mc_T37+mc_T36+mc_T29;
    mc_T43 = -1.*mc_T9*mc_T20*x*y*mc_T28;
    mc_T44 = 0.75*mc_T35*mc_T31*x*y*mc_T28;
    mc_T45 = 0.5*mc_T9*mc_T20*x*z*mc_T18;
    mc_T46 = 1+mc_T6;
    mc_T47 = 0.5*mc_T9*mc_T20*y;
    mc_T48 = mc_T47/mc_T46;
    mc_T49 = mc_T48+mc_T45+mc_T44+mc_T43;
    mc_T50.SetShape(3 );
    mc_T50(1) = mc_T49;
    mc_T50(2) = mc_T42;
    mc_T50(3) = mc_T26;

    mc_T51 = 0.+mc_T24;
    mc_T52 = -1.*mc_T9*mc_T20*mc_T6*mc_T28;
    mc_T53 = 0.75*mc_T35*mc_T31*mc_T6*mc_T28;
    mc_T54 = 1.*mc_T9*mc_T20*x;
    mc_T55 = mc_T54/mc_T46;
    mc_T56 = pow(mc_T46, -2);
    mc_T57 = -2*mc_T12*x*mc_T56;
    mc_T58 = mc_T57+mc_T55+mc_T37+mc_T53+mc_T52;
    mc_T59.SetShape(3 );
    mc_T59(1) = mc_T58;
    mc_T59(2) = mc_T49;
    mc_T59(3) = mc_T51;

    mc_T60.SetShape(3, 3 );
    mc_T60.Set(1,All,mc_T59);
    mc_T60.Set(2,All,mc_T50);
    mc_T60.Set(3,All,mc_T25);

    return mc_T60;
}
\
\>", "Print",
 CellChangeTimes->{3.5003728491832*^9, 3.500372981258789*^9}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["A Large Example", "Section"],

Cell[CellGroupData[{

Cell["A New Expression", "Subsection"],

Cell["We now consider an expression with 10 parameters", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"expr", "=", 
   FractionBox[
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"Sin", "[", 
        RowBox[{"y", " ", "z"}], "]"}], "+", 
       RowBox[{"ArcTan", "[", "x", "]"}]}], ")"}], 
     RowBox[{"(", 
      RowBox[{"1", "+", 
       RowBox[{"Log", "[", 
        RowBox[{"u", " ", "v"}], "]"}]}], ")"}], 
     SuperscriptBox["\[ExponentialE]", 
      RowBox[{".5", "d"}]]}], 
    RowBox[{
     SqrtBox[
      RowBox[{"1", "+", 
       RowBox[{".5", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["x", "2"]}], "-", 
          SuperscriptBox["y", "2"]}]]}]}]], 
     RowBox[{"(", 
      RowBox[{"2", "+", 
       SuperscriptBox["v", 
        RowBox[{
         RowBox[{"-", "1"}], "/", "w"}]], "+", 
       SqrtBox[
        RowBox[{"2", "+", 
         RowBox[{"Sinh", "[", 
          RowBox[{"1", "/", "a"}], "]"}], "+", 
         RowBox[{"b", "/", "c"}]}]]}], ")"}]}]]}], ";"}]], "Input"],

Cell[TextData[{
 "We redefine the function that evaluates the Hessian of ",
 StyleBox["expr",
  FontWeight->"Bold"]
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"SymbolicMatrix", "[", 
    RowBox[{
    "x_", ",", "y_", ",", "z_", ",", "u_", ",", "v_", ",", "w_", ",", "a_", 
     ",", "b_", ",", "c_", ",", "d_"}], "]"}], "=", 
   RowBox[{"HessianMC", "[", 
    RowBox[{"expr", ",", 
     RowBox[{"{", 
      RowBox[{
      "x", ",", "y", ",", "z", ",", "u", ",", "v", ",", "w", ",", "a", ",", 
       "b", ",", "c", ",", "d"}], "}"}]}], "]"}]}], ";"}]], "Input"],

Cell["A numerical evaluation", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SymbolicMatrix", "[", 
   RowBox[{
   "1.", ",", "2.", ",", "3.", ",", "4.", ",", "5.", ",", "6.", ",", "7.", 
    ",", "8.", ",", "9.", ",", "10."}], "]"}], "//", "MatrixForm"}]], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"-", "65.47174020175036`"}], "0.8235131883092613`", 
      "0.8471438070691143`", "4.124095820660632`", "3.6721594663804566`", 
      RowBox[{"-", "0.5001097758922378`"}], "0.08659601005705414`", 
      RowBox[{"-", "0.4666968486124472`"}], "0.4148416432110642`", 
      "32.95756553968564`"},
     {"0.8235131883092613`", "333.92551643358007`", "348.1092369064476`", 
      "23.706480848329072`", "21.108621585764602`", 
      RowBox[{"-", "2.8747738509994925`"}], "0.497778602445376`", 
      RowBox[{"-", "2.682706800404611`"}], "2.384628267026321`", 
      "189.44950123611613`"},
     {"0.8471438070691143`", "348.1092369064476`", "146.84272584497592`", 
      "15.78569982776161`", "14.055834194140864`", 
      RowBox[{"-", "1.9142578510456079`"}], "0.33146141129754547`", 
      RowBox[{"-", "1.786364013622328`"}], "1.5878791232198468`", 
      "126.15086052484548`"},
     {"4.124095820660632`", "23.706480848329072`", "15.78569982776161`", 
      RowBox[{"-", "1.0398273334851629`"}], "0.02352927053449515`", 
      RowBox[{"-", "0.03155741670844589`"}], "0.0054642930540275475`", 
      RowBox[{"-", "0.029449028269655293`"}], "0.02617691401747137`", 
      "2.0796546669703257`"},
     {"3.6721594663804566`", "21.108621585764602`", "14.055834194140864`", 
      "0.02352927053449515`", 
      RowBox[{"-", "0.7113367198710191`"}], 
      RowBox[{"-", "0.07681761411693375`"}], "0.005359549674508734`", 
      RowBox[{"-", "0.028884528760933816`"}], "0.025675136676385612`", 
      "1.8517570648719912`"},
     {
      RowBox[{"-", "0.5001097758922378`"}], 
      RowBox[{"-", "2.8747738509994925`"}], 
      RowBox[{"-", "1.9142578510456079`"}], 
      RowBox[{"-", "0.03155741670844589`"}], 
      RowBox[{"-", "0.07681761411693375`"}], "0.1532311741140218`", 
      RowBox[{"-", "0.0013252584292386663`"}], "0.007142291337848824`", 
      RowBox[{"-", "0.006348703411421176`"}], 
      RowBox[{"-", "0.25218997682385835`"}]},
     {"0.08659601005705414`", "0.497778602445376`", "0.33146141129754547`", 
      "0.0054642930540275475`", "0.005359549674508734`", 
      RowBox[{"-", "0.0013252584292386663`"}], 
      RowBox[{"-", "0.02467949852640713`"}], 
      RowBox[{"-", "0.00283684668691948`"}], "0.0025216414994839835`", 
      "0.043667704216269546`"},
     {
      RowBox[{"-", "0.4666968486124472`"}], 
      RowBox[{"-", "2.682706800404611`"}], 
      RowBox[{"-", "1.786364013622328`"}], 
      RowBox[{"-", "0.029449028269655293`"}], 
      RowBox[{"-", "0.028884528760933816`"}], "0.007142291337848824`", 
      RowBox[{"-", "0.00283684668691948`"}], "0.01528878071760691`", 
      "0.03870794277628963`", 
      RowBox[{"-", "0.23534086536373097`"}]},
     {"0.4148416432110642`", "2.384628267026321`", "1.5878791232198468`", 
      "0.02617691401747137`", "0.025675136676385612`", 
      RowBox[{"-", "0.006348703411421176`"}], "0.0025216414994839835`", 
      "0.03870794277628963`", 
      RowBox[{"-", "0.08089414476188331`"}], "0.20919188032331643`"},
     {"32.95756553968564`", "189.44950123611613`", "126.15086052484548`", 
      "2.0796546669703257`", "1.8517570648719912`", 
      RowBox[{"-", "0.25218997682385835`"}], "0.043667704216269546`", 
      RowBox[{"-", "0.23534086536373097`"}], "0.20919188032331643`", 
      "16.619486541321013`"}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.5003728493238225`*^9, 3.500372981321288*^9}]
}, Open  ]],

Cell["\<\
The symbolic form of the matrix occupies approximately 30 pages!\
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Code Generation", "Subsection"],

Cell[TextData[{
 "We need to declare types of the argument and returnvalue of the function to \
be compiled. In this case a list of ",
 "10",
 " real parameters and a real 10",
 StyleBox[" \[Cross] ",
  FontWeight->"Bold"],
 "10 matrix."
}], "Text"],

Cell[BoxData[
 RowBox[{"Declare", "[", 
  RowBox[{
   RowBox[{"SymbolicMatrix", "[", 
    RowBox[{
     RowBox[{"Real", " ", "x_"}], ",", 
     RowBox[{"Real", " ", "y_"}], ",", 
     RowBox[{"Real", " ", "z_"}], ",", 
     RowBox[{"Real", " ", "u_"}], ",", 
     RowBox[{"Real", " ", "v_"}], ",", 
     RowBox[{"Real", " ", "w_"}], ",", 
     RowBox[{"Real", " ", "a_"}], ",", 
     RowBox[{"Real", " ", "b_"}], ",", 
     RowBox[{"Real", " ", "c_"}], ",", 
     RowBox[{"Real", " ", "d_"}]}], "]"}], "\[Rule]", 
   RowBox[{"Real", "[", 
    RowBox[{"10", ",", "10"}], "]"}]}], "]"}]], "Input"],

Cell[TextData[{
 "We generate C++ code and compile it to an executable file using the ",
 StyleBox["BuildCode",
  FontWeight->"Bold"],
 " command. The option ",
 StyleBox["EvaluateFunctions",
  FontWeight->"Bold"],
 " makes optimiizations such as common subexpression eliminations and \
symbolic evaluation of the function ",
 StyleBox["SymbolicMatrix",
  FontWeight->"Bold"],
 " before code generation."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BuildCode", "[", 
  RowBox[{"EvaluateFunctions", "\[Rule]", 
   RowBox[{"{", "SymbolicMatrix", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"Grad", ",", "HessianMC"}], "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`Grad, \
$CellContext`HessianMC}],
  Editable->False]], "Print",
 CellChangeTimes->{3.5003728519175224`*^9, 3.5003729816181574`*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"Grad", ",", "HessianMC"}], "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`Grad, \
$CellContext`HessianMC}],
  Editable->False]], "Print",
 CellChangeTimes->{3.5003728519175224`*^9, 3.5003729816962805`*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"Grad", ",", "HessianMC"}], "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`Grad, \
$CellContext`HessianMC}],
  Editable->False]], "Print",
 CellChangeTimes->{3.5003728519175224`*^9, 3.5003729817587795`*^9}],

Cell[BoxData["\<\"Successful compilation to C++: 1 function(s)\"\>"], "Print",
 CellChangeTimes->{3.5003728519175224`*^9, 3.500372991321096*^9}]
}, Open  ]],

Cell[TextData[{
 "Connect the executable to ",
 StyleBox["Mathematica",
  FontSlant->"Italic"]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"InstallCode", "[", "]"}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Global\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Global", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.5003729964303727`*^9}]
}, Open  ]],

Cell["Now the external code is used instead", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SymbolicMatrix", "[", 
   RowBox[{
   "1.", ",", "2.", ",", "3.", ",", "4.", ",", "5.", ",", "6.", ",", "7.", 
    ",", "8.", ",", "9.", ",", "10."}], "]"}], "//", "MatrixForm"}]], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"-", "65.47174020175035`"}], "0.8235131883092612`", 
      "0.8471438070691143`", "4.1240958206606315`", "3.6721594663804558`", 
      RowBox[{"-", "0.5001097758922377`"}], "0.08659601005705413`", 
      RowBox[{"-", "0.46669684861244715`"}], "0.4148416432110641`", 
      "32.95756553968563`"},
     {"0.8235131883092612`", "333.92551643358007`", "348.1092369064475`", 
      "23.706480848329065`", "21.108621585764595`", 
      RowBox[{"-", "2.8747738509994916`"}], "0.49777860244537586`", 
      RowBox[{"-", "2.68270680040461`"}], "2.38462826702632`", 
      "189.44950123611605`"},
     {"0.8471438070691143`", "348.1092369064475`", "146.8427258449759`", 
      "15.785699827761608`", "14.05583419414086`", 
      RowBox[{"-", "1.9142578510456076`"}], "0.3314614112975454`", 
      RowBox[{"-", "1.7863640136223276`"}], "1.5878791232198466`", 
      "126.15086052484546`"},
     {"4.1240958206606315`", "23.706480848329065`", "15.785699827761608`", 
      RowBox[{"-", "1.0398273334851627`"}], "0.023529270534495144`", 
      RowBox[{"-", "0.03155741670844588`"}], "0.005464293054027547`", 
      RowBox[{"-", "0.029449028269655286`"}], "0.026176914017471364`", 
      "2.0796546669703253`"},
     {"3.6721594663804558`", "21.108621585764595`", "14.05583419414086`", 
      "0.023529270534495144`", 
      RowBox[{"-", "0.7113367198710189`"}], 
      RowBox[{"-", "0.07681761411693369`"}], "0.0053595496745087335`", 
      RowBox[{"-", "0.028884528760933805`"}], "0.0256751366763856`", 
      "1.8517570648719905`"},
     {
      RowBox[{"-", "0.5001097758922377`"}], 
      RowBox[{"-", "2.8747738509994916`"}], 
      RowBox[{"-", "1.9142578510456076`"}], 
      RowBox[{"-", "0.03155741670844588`"}], 
      RowBox[{"-", "0.07681761411693369`"}], "0.15323117411402173`", 
      RowBox[{"-", "0.0013252584292386658`"}], "0.007142291337848822`", 
      RowBox[{"-", "0.006348703411421175`"}], 
      RowBox[{"-", "0.2521899768238583`"}]},
     {"0.08659601005705413`", "0.49777860244537586`", "0.3314614112975454`", 
      "0.005464293054027547`", "0.0053595496745087335`", 
      RowBox[{"-", "0.0013252584292386658`"}], 
      RowBox[{"-", "0.024679498526407127`"}], 
      RowBox[{"-", "0.00283684668691948`"}], "0.002521641499483982`", 
      "0.04366770421626954`"},
     {
      RowBox[{"-", "0.46669684861244715`"}], 
      RowBox[{"-", "2.68270680040461`"}], 
      RowBox[{"-", "1.7863640136223276`"}], 
      RowBox[{"-", "0.029449028269655286`"}], 
      RowBox[{"-", "0.028884528760933805`"}], "0.007142291337848822`", 
      RowBox[{"-", "0.00283684668691948`"}], "0.015288780717606906`", 
      "0.03870794277628962`", 
      RowBox[{"-", "0.23534086536373092`"}]},
     {"0.4148416432110641`", "2.38462826702632`", "1.5878791232198466`", 
      "0.026176914017471364`", "0.0256751366763856`", 
      RowBox[{"-", "0.006348703411421175`"}], "0.002521641499483982`", 
      "0.03870794277628962`", 
      RowBox[{"-", "0.0808941447618833`"}], "0.20919188032331637`"},
     {"32.95756553968563`", "189.44950123611605`", "126.15086052484546`", 
      "2.0796546669703253`", "1.8517570648719905`", 
      RowBox[{"-", "0.2521899768238583`"}], "0.04366770421626954`", 
      RowBox[{"-", "0.23534086536373092`"}], "0.20919188032331637`", 
      "16.61948654132101`"}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.500372859870495*^9, 3.5003729964459977`*^9}]
}, Open  ]],

Cell[TextData[{
 "Remove the link to the external code (",
 StyleBox["Mathematicas",
  FontSlant->"Italic"],
 " own definition of ",
 StyleBox["SymbolicMatrix",
  FontWeight->"Bold"],
 " is used again)."
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"UninstallCode", "[", "]"}], ";"}]], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["A Performance Comparision", "Subsection"],

Cell[TextData[{
 "To compare performance between compiled and uncompiled code we define a \
function that uses ",
 StyleBox["SymbolicMatrix",
  FontWeight->"Bold"],
 " a specified number of times."
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RepeatComputation", "[", 
    RowBox[{"Integer", " ", "n_"}], "]"}], "\[Rule]", "Null"}], ":=", 
  RowBox[{"Do", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"SymbolicMatrix", "[", 
      RowBox[{
       RowBox[{"1.", "+", 
        RowBox[{"1", "/", "n"}]}], ",", 
       RowBox[{"2.", "+", 
        RowBox[{"1", "/", "n"}]}], ",", 
       RowBox[{"3.", "+", 
        RowBox[{"1", "/", "n"}]}], ",", 
       RowBox[{"4.", "+", 
        RowBox[{"1", "/", "n"}]}], ",", 
       RowBox[{"5.", "+", 
        RowBox[{"1", "/", "n"}]}], ",", 
       RowBox[{"6.", "+", 
        RowBox[{"1", "/", "n"}]}], ",", 
       RowBox[{"7.", "+", 
        RowBox[{"1", "/", "n"}]}], ",", 
       RowBox[{"8.", "+", 
        RowBox[{"1", "/", "n"}]}], ",", 
       RowBox[{"9.", "+", 
        RowBox[{"1", "/", "n"}]}], ",", 
       RowBox[{"10.", "+", 
        RowBox[{"1", "/", "n"}]}]}], "]"}], ";"}], ",", 
    RowBox[{"{", "n", "}"}]}], "]"}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BuildCode", "[", 
  RowBox[{"EvaluateFunctions", "\[Rule]", 
   RowBox[{"{", "SymbolicMatrix", "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"Grad", ",", "HessianMC"}], "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`Grad, \
$CellContext`HessianMC}],
  Editable->False]], "Print",
 CellChangeTimes->{3.500372862401696*^9, 3.5003729967272425`*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"Grad", ",", "HessianMC"}], "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`Grad, \
$CellContext`HessianMC}],
  Editable->False]], "Print",
 CellChangeTimes->{3.500372862401696*^9, 3.5003729968053656`*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"Grad", ",", "HessianMC"}], "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`Grad, \
$CellContext`HessianMC}],
  Editable->False]], "Print",
 CellChangeTimes->{3.500372862401696*^9, 3.5003729968678646`*^9}],

Cell[BoxData["\<\"Successful compilation to C++: 2 function(s)\"\>"], "Print",
 CellChangeTimes->{3.500372862401696*^9, 3.5003730061645613`*^9}]
}, Open  ]],

Cell["\<\
We do 100 iterations with the uncompiled code and compute the average time:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"n", "=", "100"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"start", "=", 
   RowBox[{"AbsoluteTime", "[", "]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"RepeatComputation", "[", "n", "]"}], ";"}], "\n", 
 RowBox[{"time1", "=", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"AbsoluteTime", "[", "]"}], "-", "start"}], ")"}], "/", 
   "n"}]}]}], "Input"],

Cell[BoxData["0.004843657`7.136718374812255"], "Output",
 CellChangeTimes->{3.5003728630735583`*^9, 3.5003730118519516`*^9}]
}, Open  ]],

Cell[TextData[{
 "Install the compiled code into ",
 StyleBox["Mathematica",
  FontSlant->"Italic"]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"InstallCode", "[", "]"}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Global\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Global", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.5003730120394483`*^9}]
}, Open  ]],

Cell["\<\
We do 100000 iterations with the uncompiled code and compute the average time:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"n", "=", "100000"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"start", "=", 
   RowBox[{"AbsoluteTime", "[", "]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"RepeatComputation", "[", "n", "]"}], ";"}], "\n", 
 RowBox[{"time2", "=", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"AbsoluteTime", "[", "]"}], "-", "start"}], ")"}], "/", 
   "n"}]}]}], "Input"],

Cell[BoxData["8.12484400000000000000283`7.361360024612783*^-6"], "Output",
 CellChangeTimes->{3.5003730128519325`*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"UninstallCode", "[", "]"}], ";"}]], "Input"],

Cell["Speed-up:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"time1", "/", "time2"}]], "Input"],

Cell[BoxData["596.1538461538461561698`6.933643626070981"], "Output",
 CellChangeTimes->{3.5003730129144316`*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean Up", "Subsection"],

Cell[TextData[{
 "Remove the definition of ",
 StyleBox["SymbolicMatrix",
  FontWeight->"Bold"],
 " and all generated files."
}], "Text"],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{"SymbolicMatrix", ",", "RepeatComputation"}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"CleanMathCodeFiles", "[", 
  RowBox[{
   RowBox[{"Confirm", "\[Rule]", "False"}], ",", 
   RowBox[{"CleanAllBut", "\[Rule]", 
    RowBox[{"{", "}"}]}]}], "]"}]], "Input"]
}, Open  ]]
}, Open  ]]
},
AutoGeneratedPackage->Automatic,
ScreenStyleEnvironment->"Working",
WindowToolbars->{},
WindowSize->{1272, 922},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
PrintingCopies->1,
PrintingPageRange->{Automatic, Automatic},
FrontEndVersion->"6.0 for Microsoft Windows (32-bit) (April 20, 2007)",
StyleDefinitions->Notebook[{
   Cell[
    CellGroupData[{
      Cell["Style Definitions", "Subtitle"], 
      Cell[
      "Modify the definitions below to change the default appearance of all \
cells in a given style.  Make modifications to any definition using commands \
in the Format menu.", "Text"], 
      Cell[
       CellGroupData[{
         Cell["Style Environment Names", "Section"], 
         Cell[
          StyleData[All, "Working"], PageWidth -> WindowWidth, ScriptMinSize -> 
          9], 
         Cell[
          StyleData[All, "Presentation"], PageWidth -> WindowWidth, 
          ScriptMinSize -> 12, FontSize -> 16], 
         Cell[
          StyleData[All, "Condensed"], PageWidth -> WindowWidth, 
          CellBracketOptions -> {"Margins" -> {1, 1}, "Widths" -> {0, 5}}, 
          ScriptMinSize -> 8, FontSize -> 11], 
         Cell[
          StyleData[All, "Printout"], PageWidth -> PaperWidth, ScriptMinSize -> 
          7, FontSize -> 10, 
          PrivateFontOptions -> {"FontType" -> "Outline"}]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Notebook Options", "Section"], 
         Cell[
         "The options defined for the style below will be used at the \
Notebook level.", "Text"], 
         Cell[
          StyleData["Notebook"], PageHeaders -> {{
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"], None, 
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"]}, {
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"], None, 
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"]}}, 
          PageHeaderLines -> {True, True}, 
          PrintingOptions -> {
           "FacingPages" -> True, "FirstPageHeader" -> False}, 
          CellFrameLabelMargins -> 6, StyleMenuListing -> None]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headings", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], ShowCellBracket -> False, 
             CellMargins -> {{0, 0}, {0, 0}}, PageBreakBelow -> False, 
             LineSpacing -> {1, 0}, CounterIncrements -> "Title", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subtitle", 0}, {"Subsubtitle", 0}}, FontSize -> 
             34, FontColor -> GrayLevel[1], Background -> RGBColor[0, 0, 1]], 
            
            Cell[
             StyleData["Title", "Presentation"], 
             CellMargins -> {{0, 0}, {0, 0}}, LineSpacing -> {1, 0}, FontSize -> 
             44], 
            Cell[
             StyleData["Title", "Condensed"], CellMargins -> {{0, 0}, {0, 0}},
              FontSize -> 20], 
            Cell[
             StyleData["Title", "Printout"], CellMargins -> {{0, 0}, {0, 0}}, 
             FontSize -> 24, FontTracking -> "Plain", Background -> 
             GrayLevel[0]]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], ShowCellBracket -> False, 
             CellMargins -> {{0, 0}, {0, 0}}, PageBreakBelow -> False, 
             LineSpacing -> {1, 3}, ParagraphIndent -> -96, CounterIncrements -> 
             "Subtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subsubtitle", 0}}, FontFamily -> "Helvetica", 
             FontSize -> 18, FontColor -> GrayLevel[1], Background -> 
             RGBColor[0.2, 0.700008, 0.700008]], 
            Cell[
             StyleData["Subtitle", "Presentation"], 
             CellMargins -> {{0, 0}, {0, 0}}, LineSpacing -> {1, 6}, 
             ParagraphIndent -> -157, FontSize -> 30], 
            Cell[
             StyleData["Subtitle", "Condensed"], 
             CellMargins -> {{0, 0}, {0, 0}}, ParagraphIndent -> -78, 
             FontSize -> 14], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{0, 0}, {0, 0}}, ParagraphIndent -> -85, 
             FontSize -> 16, Background -> GrayLevel[0.6]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], ShowCellBracket -> False, 
             CellMargins -> {{10, 4}, {30, 10}}, PageBreakBelow -> False, 
             CounterIncrements -> "Subsubtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}}, FontFamily -> "Helvetica", FontSize -> 14, 
             FontSlant -> "Italic"], 
            Cell[
             StyleData["Subsubtitle", "Presentation"], 
             CellMargins -> {{8, 10}, {40, 20}}, LineSpacing -> {1, 0}, 
             FontSize -> 24], 
            Cell[
             StyleData["Subsubtitle", "Condensed"], 
             CellMargins -> {{8, 10}, {12, 8}}, FontSize -> 12], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{9, 10}, {50, 10}}, FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], CellFrame -> {{6, 0}, {0, 1}}, CellDingbat -> 
             None, CellMargins -> {{12, Inherited}, {4, 24}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameMargins -> 6, CounterIncrements -> "Section", 
             CounterAssignments -> {{"Subsection", 0}, {"Subsubsection", 0}}, 
             FontFamily -> "Helvetica", FontSize -> 16, FontWeight -> "Bold", 
             FontColor -> RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Section", "Presentation"], 
             CellMargins -> {{10, 10}, {8, 32}}, LineSpacing -> {1, 2}, 
             FontSize -> 24, FontTracking -> "Condensed"], 
            Cell[
             StyleData["Section", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Section", "Printout"], 
             CellMargins -> {{9, 0}, {2, 50}}, FontSize -> 14, FontTracking -> 
             "Plain", FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], 
             CellMargins -> {{12, Inherited}, {8, 20}}, 
             CellGroupingRules -> {"SectionGrouping", 40}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsection", "Presentation"], 
             CellMargins -> {{11, 10}, {8, 32}}, LineSpacing -> {1, 0}, 
             FontSize -> 22], 
            Cell[
             StyleData["Subsection", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Subsection", "Printout"], 
             CellMargins -> {{9, 0}, {4, 40}}, FontSize -> 12]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> "\[FilledSquare]", 
             CellMargins -> {{25, Inherited}, {8, 12}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsubsection", FontFamily -> 
             "Times", FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsubsection", "Presentation"], 
             CellMargins -> {{29, 10}, {8, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Subsubsection", "Condensed"], 
             CellMargins -> {{22, Inherited}, {2, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Subsubsection", "Printout"], 
             CellMargins -> {{21, 0}, {4, 20}}, FontSize -> 11]}, Open]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Body Text", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Text"], CellMargins -> {{12, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 12}, 
             CounterIncrements -> "Text", FontFamily -> "Times"], 
            Cell[
             StyleData["Text", "Presentation"], 
             CellMargins -> {{13, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["Text", "Condensed"], CellMargins -> {{8, 10}, {4, 4}},
              LineSpacing -> {1, 1}, ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["Text", "Printout"], CellMargins -> {{9, 0}, {4, 4}}, 
             ParagraphSpacing -> {0, 6}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SmallText"], CellMargins -> {{12, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 6}, 
             CounterIncrements -> "SmallText", FontFamily -> "Helvetica", 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Presentation"], 
             CellMargins -> {{13, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["SmallText", "Condensed"], 
             CellMargins -> {{8, 10}, {2, 2}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Printout"], 
             CellMargins -> {{9, 0}, {4, 4}}, FontSize -> 7]}, Closed]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Input/Output", "Section"], 
         Cell[
         "The cells in this section define styles used for input and output \
to the kernel.  Be careful when modifying, renaming, or removing these \
styles, because the front end associates special meanings with these style \
names.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Input"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, Evaluatable -> True, 
             CellGroupingRules -> "InputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{5, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, FormatType -> 
             InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
             CounterIncrements -> "Input", FontWeight -> "Bold", Background -> 
             RGBColor[1, 1, 0]], 
            Cell[
             StyleData["Input", "Presentation"], 
             CellMargins -> {{62, Inherited}, {10, 10}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Input", "Condensed"], 
             CellMargins -> {{40, 10}, {4, 4}}], 
            Cell[
             StyleData["Input", "Printout"], CellMargins -> {{44, 0}, {6, 6}},
              Background -> GrayLevel[0.8]]}, Closed]], 
         Cell[
          StyleData["InlineInput"], Evaluatable -> True, CellGroupingRules -> 
          "InputGrouping", CellHorizontalScrolling -> True, PageBreakWithin -> 
          False, GroupPageBreakWithin -> False, DefaultFormatType -> 
          DefaultInputFormatType, AutoItalicWords -> {}, FormatType -> 
          InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
          CounterIncrements -> "Input", FontWeight -> "Bold"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Output"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, CellEditDuplicate -> True, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, 
             CellLabelMargins -> {{3, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Output", Background -> 
             RGBColor[1, 0.900008, 0.900008]], 
            Cell[
             StyleData["Output", "Presentation"], 
             CellMargins -> {{62, Inherited}, {12, 5}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Output", "Condensed"], 
             CellMargins -> {{40, Inherited}, {4, 1}}], 
            Cell[
             StyleData["Output", "Printout"], 
             CellMargins -> {{44, 0}, {6, 2}}, Background -> 
             GrayLevel[0.900008]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Message"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Message", StyleMenuListing -> 
             None, FontColor -> RGBColor[1, 0, 0]], 
            Cell[
             StyleData["Message", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Message", "Condensed"], 
             CellMargins -> {{50, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Message", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}, 
             FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Print"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, ShowCellLabel -> 
             False, DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Print", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Print", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Print", "Condensed"], 
             CellMargins -> {{50, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Print", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Graphics"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "GraphicsGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Graphics", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Graphics", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Graphics", "Condensed"], 
             CellMargins -> {{52, Inherited}, {Inherited, Inherited}}, 
             ImageSize -> {175, 175}], 
            Cell[
             StyleData["Graphics", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}, 
             ImageSize -> {250, 250}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["CellLabel"], StyleMenuListing -> None, FontFamily -> 
             "Helvetica", FontSize -> 11, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["CellLabel", "Presentation"], FontSize -> 12], 
            Cell[
             StyleData["CellLabel", "Condensed"], FontSize -> 8], 
            Cell[
             StyleData["CellLabel", "Printout"], FontSize -> 8, FontColor -> 
             GrayLevel[0]]}, Closed]]}, Open]], 
      Cell[
       CellGroupData[{
         Cell["Unique Styles", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], ShowCellBracket -> False, 
             CellMargins -> {{10, 4}, {2, 10}}, FontSize -> 16, FontSlant -> 
             "Italic"], 
            Cell[
             StyleData["Author", "Presentation"], 
             CellMargins -> {{12, 10}, {2, 12}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}, FontSize -> 20], 
            Cell[
             StyleData["Author", "Condensed"], 
             CellMargins -> {{8, 10}, {1, 4}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}, FontSize -> 12], 
            Cell[
             StyleData["Author", "Printout"], 
             CellMargins -> {{9, 0}, {4, 12}}, ParagraphSpacing -> {0, 6}, 
             FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Copyright"], ShowCellBracket -> False, 
             CellMargins -> {{10, 10}, {40, 2}}, FontFamily -> "Helvetica", 
             FontSize -> 9], 
            Cell[
             StyleData["Copyright", "Presentation"], 
             CellMargins -> {{12, 10}, {50, 2}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["Copyright", "Condensed"], 
             CellMargins -> {{8, 10}, {12, 1}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["Copyright", "Printout"], 
             CellMargins -> {{9, 0}, {72, 4}}, FontSize -> 7]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Example"], CellMargins -> {{12, 10}, {5, 12}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 12}, 
             CounterIncrements -> "Example", FontFamily -> "Times", 
             FontWeight -> "Bold"], 
            Cell[
             StyleData["Example", "Presentation"], 
             CellMargins -> {{18, 10}, {8, 20}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["Example", "Condensed"], 
             CellMargins -> {{8, 10}, {4, 8}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["Example", "Printout"], 
             CellMargins -> {{9, 0}, {4, 10}}, ParagraphSpacing -> {0, 6}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Definition"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{23, Inherited}, {Inherited, Inherited}}, 
             ShowStringCharacters -> True, CounterIncrements -> "Definition", 
             FontFamily -> "Helvetica", FontWeight -> "Bold", FontColor -> 
             GrayLevel[1], Background -> RGBColor[0.2, 0.700008, 0.700008]], 
            Cell[
             StyleData["Definition", "Presentation"], 
             CellMargins -> {{62, Inherited}, {5, 12}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Definition", "Condensed"], 
             CellMargins -> {{40, 10}, {1, 4}}], 
            Cell[
             StyleData["Definition", "Printout"], 
             CellMargins -> {{44, 0}, {2, 6}}, Background -> GrayLevel[0.6]]},
            Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Theorem"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{23, Inherited}, {Inherited, Inherited}}, 
             ShowStringCharacters -> True, CounterIncrements -> "Theorem", 
             FontFamily -> "Helvetica", FontWeight -> "Bold", FontColor -> 
             GrayLevel[1], Background -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Theorem", "Presentation"], 
             CellMargins -> {{62, Inherited}, {5, 12}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Theorem", "Condensed"], 
             CellMargins -> {{40, 10}, {1, 4}}], 
            Cell[
             StyleData["Theorem", "Printout"], 
             CellMargins -> {{44, 0}, {2, 6}}, Background -> GrayLevel[0.4]]},
            Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ExerciseMain"], CellFrame -> {{6, 0}, {0, 1}}, 
             CellMargins -> {{12, Inherited}, {4, 20}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameColor -> RGBColor[0.571389, 0.19675, 0.570504], 
             CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 16, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["ExerciseMain", "Presentation"], 
             CellMargins -> {{18, 10}, {8, 32}}, LineSpacing -> {1, 2}, 
             FontSize -> 24, FontTracking -> "Condensed"], 
            Cell[
             StyleData["ExerciseMain", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["ExerciseMain", "Printout"], 
             CellMargins -> {{9, 0}, {2, 50}}, CellFrameColor -> 
             GrayLevel[0.500008], FontSize -> 14, FontTracking -> "Plain", 
             FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Exercise"], CellDingbat -> "\[FilledDownTriangle]", 
             CellMargins -> {{23, Inherited}, {4, 18}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsubsection", FontFamily -> 
             "Times", FontSize -> 13, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Exercise", "Presentation"], 
             CellMargins -> {{33, 10}, {8, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Exercise", "Condensed"], 
             CellMargins -> {{17, Inherited}, {2, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Exercise", "Printout"], 
             CellFrame -> {{0, 0}, {0.5, 0}}, CellDingbat -> None, 
             CellMargins -> {{9, 0}, {6, 20}}, FontSize -> 11, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ExerciseText"], CellMargins -> {{62, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 8}, 
             CounterIncrements -> "ExcersiceText", FontFamily -> "Times"], 
            Cell[
             StyleData["ExerciseText", "Presentation"], 
             CellMargins -> {{74, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["ExerciseText", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 2}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["ExerciseText", "Printout"], 
             CellMargins -> {{54, 0}, {4, 4}}, ParagraphSpacing -> {0, 6}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Itemize"], CellDingbat -> "\[FilledSmallCircle]", 
             CellMargins -> {{65, Inherited}, {8, 8}}, FontFamily -> "Times", 
             FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Itemize", "Presentation"], 
             CellMargins -> {{Inherited, 10}, {8, 12}}, LineSpacing -> {1, 0},
              FontSize -> 18], 
            Cell[
             StyleData["Itemize", "Condensed"], 
             CellMargins -> {{Inherited, Inherited}, {2, 12}}, FontSize -> 
             10], 
            Cell[
             StyleData["Itemize", "Printout"], 
             CellMargins -> {{Inherited, 0}, {4, 20}}, FontSize -> 11]}, 
           Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayCode"], CellFrame -> 2, 
             CellMargins -> {{52, 10}, {8, 8}}, CellHorizontalScrolling -> 
             True, PageBreakWithin -> True, 
             CellLabelMargins -> {{5, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, FormatType -> 
             InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
             FontWeight -> "Bold", Background -> RGBColor[1, 1, 0]], 
            Cell[
             StyleData["DisplayCode", "Presentation"], 
             CellMargins -> {{62, Inherited}, {10, 10}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["DisplayCode", "Condensed"], 
             CellMargins -> {{40, 10}, {4, 4}}], 
            Cell[
             StyleData["DisplayCode", "Printout"], 
             CellMargins -> {{44, 0}, {6, 6}}, Background -> GrayLevel[0.8]]},
            Closed]]}, Open]], 
      Cell[
       CellGroupData[{
         Cell["Formulas and Programming", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayFormula"], CellMargins -> {{62, 10}, {2, 10}}, 
             CellHorizontalScrolling -> True, ScriptLevel -> 0, 
             SingleLetterItalics -> True, 
             UnderoverscriptBoxOptions -> {LimitsPositioning -> True}], 
            Cell[
             StyleData["DisplayFormula", "Presentation"], 
             CellMargins -> {{74, 10}, {2, 10}}, FontSize -> 18], 
            Cell[
             StyleData["DisplayFormula", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["DisplayFormula", "Printout"], 
             CellMargins -> {{54, 10}, {2, 10}}, FontSize -> 10]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ChemicalFormula"], CellMargins -> {{62, 10}, {2, 10}},
              AutoSpacing -> False, ScriptBaselineShifts -> {0.6, Automatic}, 
             ScriptLevel -> 1, SingleLetterItalics -> False, ZeroWidthTimes -> 
             True], 
            Cell[
             StyleData["ChemicalFormula", "Presentation"], 
             CellMargins -> {{74, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["ChemicalFormula", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["ChemicalFormula", "Printout"], 
             CellMargins -> {{54, 10}, {2, 10}}, FontSize -> 10]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Program"], CellMargins -> {{12, 10}, {Inherited, 6}}, 
             FontFamily -> "Courier"], 
            Cell[
             StyleData["Program", "Presentation"], 
             CellMargins -> {{13, 30}, {Inherited, 4}}, FontSize -> 9.5], 
            Cell[
             StyleData["Program", "Condensed"], 
             CellMargins -> {{8, 10}, {Inherited, 4}}, FontSize -> 9.5], 
            Cell[
             StyleData["Program", "Printout"], 
             CellMargins -> {{9, 0}, {Inherited, 4}}, FontSize -> 9.5]}, 
           Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Automatic Numbering", "Section"], 
         Cell[
         "The following styles are useful for numbered equations, figures, \
etc.  They automatically give the cell a FrameLabel containing a reference to \
a particular counter, and also increment that counter.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedEquation"], 
             CellMargins -> {{62, 10}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, 
                Cell[
                 TextData[{"(", 
                   CounterBox["NumberedEquation"], ")"}]]}, {None, None}}, 
             DefaultFormatType -> DefaultInputFormatType, CounterIncrements -> 
             "NumberedEquation", FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedEquation", "Presentation"], 
             CellMargins -> {{74, 10}, {Inherited, Inherited}}], 
            Cell[
             StyleData["NumberedEquation", "Condensed"], 
             CellMargins -> {{52, 10}, {Inherited, Inherited}}], 
            Cell[
             StyleData["NumberedEquation", "Printout"], 
             CellMargins -> {{54, 0}, {Inherited, Inherited}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedFigure"], 
             CellMargins -> {{62, 145}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, None}, {
                Cell[
                 TextData[{"Figure ", 
                   CounterBox["NumberedFigure"]}]], None}}, CounterIncrements -> 
             "NumberedFigure", 
             ImageMargins -> {{43, Inherited}, {Inherited, 0}}, 
             FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedFigure", "Presentation"]], 
            Cell[
             StyleData["NumberedFigure", "Condensed"]], 
            Cell[
             StyleData["NumberedFigure", "Printout"]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedTable"], 
             CellMargins -> {{62, 145}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, None}, {
                Cell[
                 TextData[{"Table ", 
                   CounterBox["NumberedTable"]}]], None}}, TextAlignment -> 
             Center, CounterIncrements -> "NumberedTable", 
             FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedTable", "Presentation"]], 
            Cell[
             StyleData["NumberedTable", "Condensed"]], 
            Cell[
             StyleData["NumberedTable", "Printout"]]}, Open]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headers and Footers", "Section"], 
         Cell[
          StyleData["Header"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 9, 
          FontSlant -> "Italic"], 
         Cell[
          StyleData["Footer"], CellMargins -> {{0, 0}, {0, 4}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 6],
          
         Cell[
          StyleData["PageNumber"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 9, 
          FontWeight -> "Bold"]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Hyperlink Styles", "Section"], 
         Cell[
         "The cells below define styles useful for making hypertext \
ButtonBoxes.  The \"Hyperlink\" style is for links within the same Notebook, \
or between Notebooks.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Hyperlink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`NotebookLocate[#2]}]& ), ButtonNote -> 
               ButtonData}], 
            Cell[
             StyleData["Hyperlink", "Presentation"]], 
            Cell[
             StyleData["Hyperlink", "Condensed"]], 
            Cell[
             StyleData["Hyperlink", "Printout"], FontColor -> GrayLevel[0], 
             Background -> GrayLevel[1]]}, Closed]], 
         Cell[
         "The following styles are for linking automatically to the on-line \
help system.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["MainBookLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["MainBook", #]}]& )}], 
            Cell[
             StyleData["MainBookLink", "Presentation"]], 
            Cell[
             StyleData["MainBookLink", "Condensed"]], 
            Cell[
             StyleData["MainBookLink", "Printout"], FontColor -> GrayLevel[0],
              Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["AddOnsLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> GrayLevel[1], Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["AddOns", #]}]& )}], 
            Cell[
             StyleData["AddOnsLink", "Presentation"]], 
            Cell[
             StyleData["AddOnsLink", "Condensed"]], 
            Cell[
             StyleData["AddOnLink", "Printout"], FontColor -> GrayLevel[0], 
             Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["RefGuideLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> GrayLevel[1], Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["RefGuideLink", #]}]& )}], 
            Cell[
             StyleData["RefGuideLink", "Presentation"]], 
            Cell[
             StyleData["RefGuideLink", "Condensed"]], 
            Cell[
             StyleData["RefGuideLink", "Printout"], FontColor -> GrayLevel[0],
              Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["GettingStartedLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["GettingStarted", #]}]& )}], 
            Cell[
             StyleData["GettingStartedLink", "Presentation"]], 
            Cell[
             StyleData["GettingStartedLink", "Condensed"]], 
            Cell[
             StyleData["GettingStartedLink", "Printout"], FontColor -> 
             GrayLevel[0], Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["OtherInformationLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["OtherInformation", #]}]& )}], 
            Cell[
             StyleData["OtherInformationLink", "Presentation"]], 
            Cell[
             StyleData["OtherInformationLink", "Condensed"]], 
            Cell[
             StyleData["OtherInformationLink", "Printout"], FontColor -> 
             GrayLevel[0], Background -> GrayLevel[1]]}, Closed]]}, Closed]]},
      Open]]}, Visible -> False, FrontEndVersion -> 
  "6.0 for Microsoft Windows (32-bit) (April 20, 2007)", StyleDefinitions -> 
  "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[580, 21, 54, 0, 55, "Title"],
Cell[637, 23, 141, 3, 54, "Copyright"],
Cell[781, 28, 384, 8, 43, "Text"],
Cell[CellGroupData[{
Cell[1190, 40, 33, 0, 54, "Section"],
Cell[CellGroupData[{
Cell[1248, 44, 35, 0, 44, "Subsection"],
Cell[1286, 46, 115, 3, 25, "Text"],
Cell[1404, 51, 495, 17, 68, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1936, 73, 59, 0, 44, "Subsection"],
Cell[1998, 75, 136, 4, 25, "Text"],
Cell[2137, 81, 191, 5, 48, "Input"],
Cell[2331, 88, 213, 6, 48, "Input"],
Cell[2547, 96, 252, 9, 48, "Input"],
Cell[CellGroupData[{
Cell[2824, 109, 142, 4, 48, "Input"],
Cell[2969, 115, 2744, 88, 83, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5762, 209, 58, 0, 44, "Subsection"],
Cell[5823, 211, 265, 8, 25, "Text"],
Cell[6091, 221, 319, 10, 48, "Input"],
Cell[CellGroupData[{
Cell[6435, 235, 147, 4, 48, "Input"],
Cell[6585, 241, 17963, 536, 311, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[24597, 783, 39, 0, 44, "Subsection"],
Cell[24639, 785, 142, 4, 25, "Text"],
Cell[24784, 791, 341, 9, 48, "Input"],
Cell[25128, 802, 38, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[25191, 806, 137, 3, 48, "Input"],
Cell[25331, 811, 903, 22, 75, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[26283, 839, 37, 0, 44, "Subsection"],
Cell[26323, 841, 115, 5, 25, "Text"],
Cell[CellGroupData[{
Cell[26463, 850, 74, 1, 48, "Input"],
Cell[26540, 853, 476, 10, 23, "Print"]
}, Open  ]],
Cell[27031, 866, 67, 0, 48, "Input"],
Cell[27101, 868, 237, 8, 25, "Text"],
Cell[27341, 878, 313, 9, 48, "Input"],
Cell[27657, 889, 415, 12, 43, "Text"],
Cell[CellGroupData[{
Cell[28097, 905, 144, 3, 48, "Input"],
Cell[28244, 910, 426, 11, 23, "Print"],
Cell[28673, 923, 426, 11, 23, "Print"],
Cell[29102, 936, 426, 11, 23, "Print"],
Cell[29531, 949, 173, 2, 23, "Print"]
}, Open  ]],
Cell[29719, 954, 106, 4, 25, "Text"],
Cell[CellGroupData[{
Cell[29850, 962, 77, 2, 48, "Input"],
Cell[29930, 966, 226, 5, 23, "Print"]
}, Open  ]],
Cell[30171, 974, 53, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[30249, 978, 137, 3, 48, "Input"],
Cell[30389, 983, 882, 21, 75, "Output"]
}, Open  ]],
Cell[31286, 1007, 215, 8, 25, "Text"],
Cell[31504, 1017, 79, 2, 48, "Input"],
Cell[31586, 1021, 107, 4, 25, "Text"],
Cell[31696, 1027, 72, 1, 48, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[31805, 1033, 40, 0, 44, "Subsection"],
Cell[31848, 1035, 529, 12, 43, "Text"],
Cell[CellGroupData[{
Cell[32402, 1051, 78, 1, 48, "Input"],
Cell[32483, 1054, 3677, 154, 2439, "Print"]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[36221, 1215, 34, 0, 54, "Section"],
Cell[CellGroupData[{
Cell[36280, 1219, 38, 0, 44, "Subsection"],
Cell[36321, 1221, 64, 0, 25, "Text"],
Cell[36388, 1223, 1005, 34, 84, "Input"],
Cell[37396, 1259, 127, 4, 25, "Text"],
Cell[37526, 1265, 448, 12, 48, "Input"],
Cell[37977, 1279, 38, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[38040, 1283, 224, 5, 48, "Input"],
Cell[38267, 1290, 3923, 77, 195, "Output"]
}, Open  ]],
Cell[42205, 1370, 88, 2, 25, "Text"]
}, Open  ]],
Cell[CellGroupData[{
Cell[42330, 1377, 37, 0, 44, "Subsection"],
Cell[42370, 1379, 249, 8, 25, "Text"],
Cell[42622, 1389, 595, 16, 48, "Input"],
Cell[43220, 1407, 416, 12, 43, "Text"],
Cell[CellGroupData[{
Cell[43661, 1423, 144, 3, 48, "Input"],
Cell[43808, 1428, 399, 10, 23, "Print"],
Cell[44210, 1440, 399, 10, 23, "Print"],
Cell[44612, 1452, 399, 10, 23, "Print"],
Cell[45014, 1464, 144, 1, 23, "Print"]
}, Open  ]],
Cell[45173, 1468, 106, 4, 25, "Text"],
Cell[CellGroupData[{
Cell[45304, 1476, 77, 2, 48, "Input"],
Cell[45384, 1480, 228, 5, 23, "Print"]
}, Open  ]],
Cell[45627, 1488, 53, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[45705, 1492, 224, 5, 48, "Input"],
Cell[45932, 1499, 3921, 77, 195, "Output"]
}, Open  ]],
Cell[49868, 1579, 215, 8, 25, "Text"],
Cell[50086, 1589, 79, 2, 48, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[50202, 1596, 47, 0, 44, "Subsection"],
Cell[50252, 1598, 209, 6, 25, "Text"],
Cell[50464, 1606, 989, 30, 48, "Input"],
Cell[CellGroupData[{
Cell[51478, 1640, 144, 3, 48, "Input"],
Cell[51625, 1645, 397, 10, 23, "Print"],
Cell[52025, 1657, 397, 10, 23, "Print"],
Cell[52425, 1669, 397, 10, 23, "Print"],
Cell[52825, 1681, 144, 1, 23, "Print"]
}, Open  ]],
Cell[52984, 1685, 99, 2, 25, "Text"],
Cell[CellGroupData[{
Cell[53108, 1691, 387, 13, 109, "Input"],
Cell[53498, 1706, 124, 1, 47, "Output"]
}, Open  ]],
Cell[53637, 1710, 111, 4, 25, "Text"],
Cell[CellGroupData[{
Cell[53773, 1718, 77, 2, 48, "Input"],
Cell[53853, 1722, 228, 5, 23, "Print"]
}, Open  ]],
Cell[54096, 1730, 102, 2, 25, "Text"],
Cell[CellGroupData[{
Cell[54223, 1736, 390, 13, 109, "Input"],
Cell[54616, 1751, 118, 1, 47, "Output"]
}, Open  ]],
Cell[54749, 1755, 79, 2, 48, "Input"],
Cell[54831, 1759, 25, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[54881, 1763, 57, 1, 48, "Input"],
Cell[54941, 1766, 112, 1, 47, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[55102, 1773, 30, 0, 44, "Subsection"],
Cell[55135, 1775, 137, 5, 25, "Text"],
Cell[55275, 1782, 111, 2, 48, "Input"],
Cell[55389, 1786, 194, 5, 48, "Input"]
}, Open  ]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
